/*--------------------------------------------------------------------------
 *
 *                         -- C N R S --
 *         Laboratoire d'Automatique et d'Analyse des Systemes
 *                 7 Avenue du colonel Roche
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloCmdTaskFunc.c
 *  Date de creation : 1994
 *
 *  Description      : Fonctions de calcul du module PILO:
 *                     Lissage de trajectoires type ligne brisse ou Reeds & Shepp.
 *                     Profiles de vitesses. Generation de consignes pour LOCO
 *
 *  Copyright (C) 1994,2009 LAAS/CNRS.
 *----------------------------------------------------------------------*/


#include <stdio.h>
#include <portLib.h>

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "h2mathLib.h"

#include "server/piloHeader.h"
#include "server/piloPosterWriteLibProto.h"
#include "piloCmdTaskCodels.h"

#include "libsmoothtraj/piloRSLibProto.h"
#include "libsmoothtraj/piloSpeedLibProto.h"
#include "libsmoothtraj/piloBrokenLineLibProto.h"

static double xtransf, ytransf, thetatransf;

/* Structure consignes calculees pour le centre du robot */
typedef struct PILO_REF_STR {
  GENPOS_MOTION_DIRECTION backFlag;
  int unused;
  double x;
  double y;
  double theta;
  double v;
  double w;
  double ds;       /* Variation curviligne (en valeur absolue) - monitors */
  double s;        /* Distance parcourue (en valeur absolue) - monitors */
}PILO_REF_STR;


/* Etat d'un mouvement. (un segment est de type PILO_TYPE_MVT) */
typedef enum {PILO_START_TRAJ,     /* Nouvelle trajectoire */
	      PILO_SEG_START,      /* Debut d'un segement */
	      PILO_SEG_ON,         /* Execution du segment */
	      PILO_SEG_END         /* Fin du segment */
} PILO_STATE_MVT;

typedef struct PILO_MON_TAB {
  double length;
}PILO_MON_TAB;

/*
 * Variables globales
 */
PILO_TRAJECTORY *piloTrajStrId = NULL;
PILO_ABSOLUTE_BL piloAbsBLStr;
SMOOTH_TRAJ *piloSmoothTrajId=NULL;


PILO_SPEED_REF_STR piloSpeedRefStr;
static PILO_SPEED_REF_STR piloSpeedRefStrSave; /* seems useless !! */
static PILO_REF_STR piloRef;             /* Consignes pour centre robot */
static PILO_STATE_MVT piloStateMvt;
static PILO_MON_TAB piloMonTab[MAX_ACTIVITIES];
static double piloCmdTaskPeriod; /* Periode */
static double piloSpeedRatio;
static int piloNbMvts;

/* Il faudrait la lire par poster ou requete */
static double distPoint = PILO_DIST_TO_CONTROLED_POINT;

static BOOL getRobotPosition (GENPOS_CART_STATE *pos);

/*****************************************************************************
 *
 *             PROTOTYPES DES FONCTIONS LOCALES
 */
static ACTIVITY_EVENT genereTraj (int *bilan);
static ACTIVITY_EVENT endExecTraj(int *bilan );
static ACTIVITY_EVENT interExecTraj(int *bilan );
static int    piloInitMvt   (int nbMvts);
static int    piloInitPos   ();
static int    piloInitGivenPos (double x, double y, double th);
static int    piloTestGivenInitPos  (double x0, double y0, double t0, double dt, double dl);
static PILO_STATE_MVT piloMoveExec  (int numMvt, int *bilan);
static double move          (double accel, double *distance, int numMvt);
static PILO_STATE_MVT piloTurnExec  (int numMvt, int *bilan);
static double turn (double accelAng, double rayon, double *angle, int numMvt);
static PILO_STATE_MVT piloClothExec (int numMvt, int *bilan);
static void   clotho        (double vitesse, double *lin, double *ang,
			     int numMvt);
static PILO_STATE_MVT piloAntiClothExec (int numMvt, int *bilan);
static void   antiClotho    (double vitesse, double *angle, int numMvt);
static int    piloMvtsFreeAndAlloc (int nbMvts);
static int    waitForLoco      (double *vRef);
static int piloWritePosterTraj(BOOL trajFlag);
static BOOL piloTestAtFinalConfig(int *bilan);

static POSTER_ID posPosterId;
static BOOL pomMEPosPosterType=FALSE;

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 *              S_pilo_POSTER_NOT_FOUND
 */

/* execInit  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
execInit(PILO_INIT_STR *initStr, int *bilan)
{
  int length;
  static  POM_ME_POS pomMEPos ;
  static  POM_POS pomPos ;

  if (strlen(initStr->posPosterName)==0) {
    printf("piloExecInit: no pos poster\n");
    return ETHER;
  }

  /* recherche du poster */
  if (posterFind(initStr->posPosterName, &posPosterId) == ERROR) {
    printf("piloExecInit: can't find posPoster\n");
    *bilan = S_pilo_POSTER_NOT_FOUND;
    return ETHER;
  }

  /* test type poster (size) */
  if (posterIoctl(posPosterId, FIO_GETSIZE, &length) == ERROR) {
    printf("piloExecInit: posterIoctl failed\n");
    posPosterId = NULL;
    *bilan = S_pilo_POSTER_NOT_FOUND;
    return ETHER;
  }
  if (length == sizeof(POM_ME_POS)) {
    printf ("piloExecInit: POM_ME_POS poster type\n");
    pomMEPosPosterType = TRUE;
  }
  else
    if (length == sizeof(POM_POS)) {
      printf ("piloExecInit: POM_POS poster type\n");
      pomMEPosPosterType = FALSE;
    }
    else {
      printf("piloExecInit: bad poster size (%d instead of %d for POM_ME_POS or of %d for POM_POS)\n",
	     length, sizeof(POM_ME_POS), sizeof(POM_POS));
      posPosterId = NULL;
      *bilan = S_pilo_POSTER_NOT_FOUND;
      return ETHER;
    }

  /* lectude de la position */
  if (pomMEPosPosterType) {
    if (piloPOM_ME_POSPosterRead(posPosterId, &pomMEPos) != OK) {
      printf ("pilo: can't read pos poster %s\n", initStr->posPosterName);
      *bilan = S_pilo_POSTER_NOT_FOUND;
      posPosterId = NULL;
      return ETHER;
    }
    else {
      pomPos.mainToBase = pomMEPos.main;
      pomPos.VLocal = pomMEPos.VLocal;
      pomPos.date = pomMEPos.date1;
    }
  } else {
    if (piloPOM_POSPosterRead(posPosterId, &pomPos) != OK) {
      printf ("pilo: can't read pos poster %s\n", initStr->posPosterName);
      posPosterId = NULL;
      *bilan = S_pilo_POSTER_NOT_FOUND;
      return ETHER;
    }
  }
  return ETHER;
}


/*****************************************************************************
 *
 *  Fonctions de demarage d'execution.
 *
 *  Parametres: pointeurs sur : Input, Output, SDI, bilan
 *
 *  Retourne: END, FAIL ou EXEC
 */

/*
 * Translation.
 *
 * Le signe de la longueur d donne le sens de deplacement.
 */

ACTIVITY_EVENT startMove (PILO_MOVE_STR *moveStr, int *bilan)
{

  /* Verifier la validite des arguments */
  if (moveStr->vPercent < 1 || moveStr->accelPercent < 1 ||
      moveStr->vPercent > 100 || moveStr->accelPercent > 100 ||
      moveStr->dL == 0.) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

  /* Allouer le tableau des mvts elementaires */
  if ((*bilan = piloMvtsFreeAndAlloc (2)) != OK)
    return ETHER;

  /* Enregistrement de la position initiale */
  if ((*bilan = piloInitPos ()) != OK)
    return ETHER;

  /* Caracteristiques du move */
  TYPE_MVT(0)  = SEGMENT;
  VAR_LIN(0) = moveStr->dL;
  RO_START(0) = 1.E10;
  C_START(0) = 0;
  V_START(0) = 0.;
  W_START(0) = 0.;

  /* Point d'arrive */
  TYPE_MVT(1) = GOTO_END;
  X_START(1) = X_START(0) + VAR_LIN(0) * cos(THETA_START(0));
  Y_START(1) = Y_START(0) + VAR_LIN(0) * sin(THETA_START(0));
  THETA_START(1) = THETA_START(0);
  RO_START(1) = 1.E10;
  C_START(1) = 0;
  V_START(1) = 0.;
  W_START(1) = 0.;

  /* Calcul des vitesses */
  memcpy(&piloSpeedRefStrSave, &piloSpeedRefStr, sizeof(PILO_SPEED_REF_STR));
  VMAX_REF = VMAX_PILO * (double) moveStr->vPercent/100.;
  ALIN_REF = AMAX_LIN_PILO * (double) moveStr->accelPercent/100.;
  printf("Valeur de Vmax et Alin : %f %f\n", VMAX_REF, ALIN_REF);

  computeSegProfiles(0, TRUE);

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(FALSE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt (2)) != OK)
    return ETHER;

  return(EXEC);
}

/*
 * Arc de cercle.
 *
 * Le signe de la variation angulaire donne le sens de rotation.
 * Le signe du rayon (= ds/dtheta) est le rpoduit des signes de translation
 * et de rotation. Autrement dit, si le centre de rotation est a` gauche
 * du robot le rayon est positif, s'il est a droite ce rayon est negatif.
 */

ACTIVITY_EVENT startTurn (PILO_TURN_STR *turnStr, int *bilan)
{
  double xRot, yRot;                 /* Centre de rotation */

  /* Verifier la validite des arguments */
  if (turnStr->wPercent < 1 || turnStr->accelPercent < 1 ||
      turnStr->wPercent > 100 || turnStr->accelPercent > 100 ||
      turnStr->dTheta == 0.) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

  /* Allouer le tableau des mvts elementaires */
  if ((*bilan = piloMvtsFreeAndAlloc (2)) != OK)
    return ETHER;

  /* Enregistrement de la position initiale */
  if ((*bilan = piloInitPos ()) != OK)
    return ETHER;

  /* Enregistrer le mouvement */
  TYPE_MVT(0) = TURN;
  VAR_ANG(0) = turnStr->dTheta;
  VAR_LIN(0) = turnStr->dTheta * turnStr->rayon;
  RO_START(0) = turnStr->rayon;
  C_START (0) = turnStr->rayon == 0 ? 1.E10 : 1/turnStr->rayon;
  W_START(0) = 0.;
  V_START(0) = 0.;

  /* Enregistrer point d'arrive' */
  xRot = X_START(0) - RO_START(0) * sin(THETA_START(0));
  yRot = Y_START(0) + RO_START(0) * cos(THETA_START(0));
  TYPE_MVT (1)   = GOTO_END;
  X_START(1)     = xRot + RO_START(0) * sin(THETA_START(0) + VAR_ANG(0));
  Y_START(1)     = yRot - RO_START(0) * cos(THETA_START(0) + VAR_ANG(0));
  THETA_START(1) = THETA_START(0) + VAR_ANG(0);
  W_START(1)     = 0.;
  V_START(1)     = 0.;
  RO_START(1)    = RO_START(0);
  C_START(1)     = C_START(0);

  /* Calcul du profile de vitesses */
  memcpy(&piloSpeedRefStrSave, &piloSpeedRefStr, sizeof(PILO_SPEED_REF_STR));
  WMAX_REF = WMAX_PILO * (double)turnStr->wPercent/100.;
  AANG_REF = AMAX_ANG_PILO * (double)turnStr->accelPercent/100.;
  VMAX_REF = VMAX_PILO;
  ALIN_REF = AMAX_LIN_PILO;
  computeTurnProfiles(0, TRUE);

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(FALSE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt (2)) != OK)
    return ETHER;

  return(EXEC);
}

ACTIVITY_EVENT startStop (int *bilan)
{
  double epsVitW = AMAX_ANG_PILO_MAJOR * piloCmdTaskPeriod;
  double epsVitV = AMAX_LIN_PILO_MAJOR * piloCmdTaskPeriod;
  double xRot, yRot;

  /* Deja a l'arret */
  if(fabs (W_START(0)) < epsVitW && fabs(V_START(0)) < epsVitV) {
    return ETHER;
  }

  /* Allouer le tableau des mvts elementaires */
  if ((*bilan = piloMvtsFreeAndAlloc (2)) != OK)
    return ETHER;

  /* Enregistrement de la position initiale */
  if ((*bilan = piloInitPos ()) != OK)
    return ETHER;

  /* Arret en arc de cercle */
  if (fabs (W_START(0)) > epsVitW) {
    TYPE_MVT(0) = TURN;
    RO_START(0) = V_START(0)/W_START(0);
    if (fabs(RO_START(0)) > 0.0001)
      ACCEL_ANG(0) = SIGN(W_START(0))
		     * MIN(AMAX_ANG_PILO_MAJOR,
			    AMAX_LIN_PILO_MAJOR/fabs(RO_START(0)));
    else
      ACCEL_ANG(0) = SIGN(W_START(0)) * AMAX_ANG_PILO_MAJOR;
    VAR_ANG(0) = SQR(W_START(0))/(2. * ACCEL_ANG(0));
    ACCEL_LIN(0) = ACCEL_ANG(0) * RO_START(0);

    /* Enregistrer point d'arrive' */
    TYPE_MVT (1) = GOTO_END;
    xRot = X_START(0) - RO_START(0) * sin(THETA_START(0));
    yRot = Y_START(0) + RO_START(0) * cos(THETA_START(0));
    X_START(1) = xRot + RO_START(0) * sin(THETA_START(0) + VAR_ANG(0));
    Y_START(1) = yRot - RO_START(0) * cos(THETA_START(0) + VAR_ANG(0));
    THETA_START(1) = THETA_START(0) + VAR_ANG(0);
  }

  /* Arret en ligne droite */
  else  {
    TYPE_MVT(0) = SEGMENT;
    RO_START(0) = SIGN(V_START(0)) * 1.E10;
    ACCEL_ANG(0) = 0.;
    ACCEL_LIN(0) = SIGN(V_START(0)) * AMAX_LIN_PILO_MAJOR;
    VAR_LIN(0) = SQR(V_START(0))/(2. * ACCEL_LIN(0));

    /* Enregistrer point d'arrive' */
    TYPE_MVT (1) = GOTO_END;
    X_START(1) = X_START(0) + VAR_LIN(0) * cos(THETA_START(0));
    Y_START(1) = Y_START(0) + VAR_LIN(0) * sin(THETA_START(0));
    THETA_START(1) = THETA_START(0);
  }

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(FALSE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt (2)) != OK)
    return ETHER;

  return(EXEC);
}

ACTIVITY_EVENT startTraj (PILO_TRAJECTORY *sm, int *bilan)
{
  int nbMvts;     /* Nombre de mouvements elementaires */

  /* Validity of parameters */
  if (sm->header.nbSegs < 1 || sm->header.nbSegs > PILO_MAX_TRAJ_SEGMENTS) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

   /* Validity of the initial position */
  if (DATA_ABS_OR_REL == PILO_ABSOLUTE_TRAJ)
    if ((*bilan = piloTestGivenInitPos(DATA_X0, DATA_Y0, DATA_THETA0,
				       DATA_DEV_THETA, DATA_DEV_L)) != OK)
      return(ETHER);

  /* Filtrage de la traj */
  switch (DATA_TYPE) {
  case PILO_LINE_SEG:
    if ((*bilan = piloSmoothFiltre(&nbMvts)) != OK) return ETHER;
    break;
  case PILO_ARC_SEG:
    if ((*bilan = piloRSFiltre(&nbMvts)) != OK) return ETHER;
    break;
  }

  /* Allouer le tableau des mvts elementaires */
  if ((*bilan = piloMvtsFreeAndAlloc (nbMvts)) != OK)
    return ETHER;

  /* Record initial position */
  if (DATA_ABS_OR_REL == PILO_RELATIVE_TRAJ) {
    if ((*bilan = piloInitPos ()) != OK)
      return ETHER;
  }
  else {
    if ((*bilan = piloInitGivenPos (DATA_X0, DATA_Y0, DATA_THETA0)) != OK)
      return ETHER;
  }

  /* Determination de la geometrie de la trajectoire */
  switch (DATA_TYPE) {
  case PILO_LINE_SEG:
    if (piloSmoothTypes (&nbMvts) != OK) {
      *bilan = S_pilo_bad_parameters;
      return ETHER;
    }
    if (piloSmoothGeo (&nbMvts) != OK) {
      *bilan = S_pilo_bad_parameters;
      return ETHER;
    }
    break;
  case PILO_ARC_SEG:
    if ((*bilan = piloRSTypes (&nbMvts)) != OK) return ETHER;
    if ((*bilan = piloRSGeo (&nbMvts)) != OK) return ETHER;
    if((*bilan = piloRSControlResult(nbMvts)) != OK) return ETHER;
    break;
  }

  /* Determination des vitesses et temps de parcour de la trajectoire */
  V_START(0) = 0.;
  W_START(0) = 0.;
  if ((*bilan = piloSpeedProfiles (nbMvts)) != OK) return ETHER;

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(TRUE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt (nbMvts)) != OK) return ETHER;

  return(EXEC);
}

ACTIVITY_EVENT startAbsSmooth (PILO_ABSOLUTE_BL *smg, int *bilan)

{
  double deltaX, deltaY;               /* Ecart entre deux consignes */
  double theta, prevTheta;             /* Angle oriente du segment */
  double dTheta;
  int nbMvts;
  int i;

  /* Validity of dynamics specifications */
  if (smg->header.nbSegs < 1 || smg->header.nbSegs > PILO_MAX_TRAJ_SEGMENTS) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

  /* Validity of the initial position */
  if (DATAG_ABS_OR_REL == PILO_ABSOLUTE_TRAJ)
    if ((*bilan = piloTestGivenInitPos(DATAG_X0, DATAG_Y0, DATAG_THETA0,
				       DATAG_DEV_THETA, DATAG_DEV_L)) != OK)
      return ETHER;

  /* Enregistrement de la position initiale */
  if (DATAG_ABS_OR_REL == PILO_RELATIVE_TRAJ) {
    if ((*bilan = piloInitPos ()) != OK)
      return ETHER;
  }
  else {
    if ((*bilan = piloInitGivenPos (DATAG_X0, DATAG_Y0, DATAG_THETA0))
	!= OK)
      return ETHER;
  }

  /**
   ** Conversion en locale
   **/
  DATA_NB = DATAG_NB;
  DATA_EPS(0) = DATAG_EPS(0);

  /*
   * 1er point
   */
  /* Longueur segment orientee */
  deltaX = DATAG_X (0) - X_START(0);
  deltaY = DATAG_Y (0) - Y_START(0);
  DATA_DL(0) = DATAG_FORW (0) * sqrt (SQR (deltaX) + SQR (deltaY));

  /* Angle segment oriente */
  theta = atan2Par (deltaY, deltaX);
  if (DATAG_FORW (0) < 0) theta = angleLimit(theta + PI);
  prevTheta = theta;

  /* Variation angulaire */
  dTheta = angleLimit(theta - THETA_START(0));

  /* Rotation la plus petite (pour le 1er point on ne peut choisir le sens) */
  DATA_DTH(0) = dTheta;
  if (fabs(dTheta) > EPSILON_ANGLE)
    printf ("PILO WARNING: initial rotation of %f rd\n", dTheta);


  /*
   * Les autres points
   */
  for (i=1; i<DATA_NB; i++) {
    deltaX = DATAG_X (i) - DATAG_X (i-1);
    deltaY = DATAG_Y (i) - DATAG_Y (i-1);
    DATA_DL(i) = DATAG_FORW(i) * sqrt (SQR (deltaX) + SQR (deltaY));

    theta = atan2Par (deltaY, deltaX);
    if (DATAG_FORW (i) < 0) theta = angleLimit(theta + PI);
    dTheta = theta - prevTheta;
    prevTheta = theta;
    /* Pour les petites variations angulaires le sens est douteux -> on prend la rotation la plus petite !?! */
    if (fabs(angleLimit(dTheta)) < 10*EPSILON_ANGLE)
      dTheta = angleLimit(dTheta);
    else if (dTheta * DATAG_TRIGO(i) < 0. ) dTheta -= SIGN(dTheta)*2.*PI;
    DATA_DTH(i) = dTheta;

    DATA_EPS(i) = DATAG_EPS(i);
  }

  /* Last rotation */
  if (DATAG_FINAL_DTHETA != 0.) {
    DATA_DTH(i) = DATAG_FINAL_DTHETA;
    DATA_DL(i) = 0.;
    DATA_EPS(i) = 0.;
    DATA_NB++;
  }

  /**
   ** Filtrage de la traj ...
   **/
  if ((*bilan = piloSmoothFiltre(&nbMvts)) != OK)
    return ETHER;

  /* Allocation du tableau de traj */
  if ((*bilan = piloMvtsFreeAndAlloc(nbMvts)) != OK)
    return ETHER;

  /* Determination de la geometrie de la trajectoire */
  if (piloSmoothTypes (&nbMvts) != OK) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }
  if (piloSmoothGeo (&nbMvts) != OK) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

  /* Determination des vitesses et temps de parcour de la trajectoire */
  V_START(0) = 0.;
  W_START(0) = 0.;
  if ((*bilan = piloSpeedProfiles (nbMvts)) != OK)
    return ETHER;

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(TRUE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt (nbMvts)) != OK)
    return ETHER;

  return(EXEC);
}


ACTIVITY_EVENT  startTrajPoster(PILO_TRAJ_POSTER  *posterName, int *bilan)
{
  POSTER_ID posterTrajId;
  int offset = sizeof(PILO_TRAJ_HEADER);
  int i;

  /* Trouver le poster */
  if (posterFind((char *) posterName, &posterTrajId) != OK)  {
    *bilan = S_pilo_POSTER_NOT_FOUND;
    return(ETHER);
  }

  /* Lecture entete */
  if (posterRead ((POSTER_ID)posterTrajId, 0, (void *) &SDI_F->trajStr.header,
		  sizeof (PILO_TRAJ_HEADER))
      != sizeof (PILO_TRAJ_HEADER)) {
    *bilan = errnoGet();
    return(ETHER);
  }

  /* Lecture de chaque point */
  for (i=0; i<SDI_F->trajStr.header.nbSegs; i++) {

    posterRead ((POSTER_ID)posterTrajId, offset,
		(void *) &(SDI_F->trajStr.seg[i]),
		sizeof(PILO_SEGMENT));

    offset += sizeof(PILO_SEGMENT);
  }

  /* Demare le traitement */
  return startTraj (&SDI_F->trajStr, bilan);
}


ACTIVITY_EVENT startMon(double *length, int *bilan)
{
  int monNum = CURRENT_ACTIVITY_NUM(PILO_CMDTASK_NUM);

  /* Initialise a la longueur deja parcourue sur la traj */
  piloMonTab[monNum].length = piloRef.s;
  return EXEC;
}

/*****************************************************************************
 *
 *  Fonctions de demarage d'execution.
 *
 *  Parametres: pointeurs sur : SDI
 *
 *  Retourne: OK ou ERROR
 */

STATUS initGenereTraj (int *bilan)
{
  piloTrajStrId = &SDI_F->trajStr;

  piloDataStrId->percentSpeed = 100;
  piloRef.v = 0.;
  piloRef.s = 0.;
  piloCmdTaskPeriod = EXEC_TASK_PERIOD(PILO_CMDTASK_NUM);

  /* Initialiser la consigne emise vers loco */
  piloDataStrId->refStr.vmax = VMAX_ROBOT;
  piloDataStrId->refStr.wmax = WMAX_ROBOT;
  piloDataStrId->refStr.linAccelMax = AMAX_LIN_ROBOT;
  piloDataStrId->refStr.angAccelMax = AMAX_ANG_ROBOT;

  piloDataStrId->speedRef.vmax = VMAX_ROBOT;
  piloDataStrId->speedRef.wmax = WMAX_ROBOT;
  piloDataStrId->speedRef.linAccelMax = AMAX_LIN_ROBOT;
  piloDataStrId->speedRef.angAccelMax = AMAX_ANG_ROBOT;

  /* Initialiser les parametres par defaut */
  SDI_F->dynamicParams.vMaxPercent = 60;
  SDI_F->dynamicParams.wMaxPercent = 60;
  SDI_F->dynamicParams.accelLinPercent = 50;
  SDI_F->dynamicParams.accelAngPercent = 50;

  /* Calcul des vitesses de consignes par defaut */
  VMAX_REF = VMAX_PILO * (double) SDI_F->dynamicParams.vMaxPercent/100.;
  WMAX_REF = WMAX_PILO * (double) SDI_F->dynamicParams.wMaxPercent/100.;
  ALIN_REF = AMAX_LIN_PILO *
	     (double) SDI_F->dynamicParams.accelLinPercent/100.;
  AANG_REF = AMAX_ANG_PILO *
	     (double) SDI_F->dynamicParams.accelAngPercent/100.;

  DATA_NB = 0;

  return(OK);
}

/*****************************************************************************
 *
 *  Fonctions d'execution.
 *
 *  Parametres: pointeurs sur : Input, Output, bilan
 *
 *  Retourne: END, FAIL ou EXEC
 */

ACTIVITY_EVENT execMove (PILO_MOVE_STR *moveStr, int *bilan)
{
  return(genereTraj(bilan));
}
ACTIVITY_EVENT execTurn (PILO_TURN_STR *turnStr, int *bilan)
{
  return(genereTraj(bilan));
}
ACTIVITY_EVENT piloGotoConfigMain(PILO_GOTO_CONFIG_STR *gotoStr, int *bilan)
{
  return(genereTraj(bilan));
}
ACTIVITY_EVENT execTraj (PILO_TRAJECTORY *smoothStr, int *bilan)
{
  return(genereTraj(bilan));
}
ACTIVITY_EVENT execAbsSmooth (PILO_ABSOLUTE_BL *smoothGlobStr, int *bilan)
{
  return(genereTraj(bilan));
}
ACTIVITY_EVENT execStop (int *bilan)
{
  return(genereTraj(bilan));
}

ACTIVITY_EVENT execMon(double *length, int *bilan)
{
  int monNum = CURRENT_ACTIVITY_NUM(PILO_CMDTASK_NUM);

  piloMonTab[monNum].length += piloRef.ds;
  if (piloMonTab[monNum].length >= *length) {
    printf("pilo mon end\n");
    return ETHER;
  }
  return EXEC;
}

ACTIVITY_EVENT execTrajPoster (PILO_TRAJ_POSTER  *posterTraj, int *bilan)
{
  return(genereTraj(bilan));
}

/*****************************************************************************
 *
 *  Fonctions en cas d'interruption
 *
 *  Parametres: pointeurs sur : Input, Output, bilan
 *
 *  Retourne: ETHER ou FAIL
 */

ACTIVITY_EVENT interMove (PILO_MOVE_STR *moveStr, int *bilan)
{
  memcpy(&piloSpeedRefStr, &piloSpeedRefStrSave, sizeof(PILO_SPEED_REF_STR));
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT interTurn (PILO_TURN_STR *turnStr, int *bilan)
{
  memcpy(&piloSpeedRefStr, &piloSpeedRefStrSave, sizeof(PILO_SPEED_REF_STR));
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT piloGotoConfigInter(PILO_GOTO_CONFIG_STR *gotoStr, int *bilan)
{
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT interTraj (PILO_TRAJECTORY *smoothStr, int *bilan)
{
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT interAbsSmooth (PILO_ABSOLUTE_BL *smoothGlobStr, int *bilan)
{
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT interStop (int *bilan)
{
  return(interExecTraj(bilan));
}
ACTIVITY_EVENT interTrajPoster (PILO_TRAJ_POSTER  *trajPoster, int *bilan)
{
  return(interExecTraj(bilan));
}

ACTIVITY_EVENT interMon (double *length, int *bilan)
{
  *bilan = S_pilo_NO_MORE_TRAJ;
  return ETHER;
}

/*****************************************************************************
 *
 *  Fonctions de fin d'execution.
 *
 *  Parametres: pointeurs sur : Input, Output, bilan
 *
 *  Retourne: ETHER ou FAIL
 */

ACTIVITY_EVENT endMove (PILO_MOVE_STR *moveStr, int *bilan)
{
  memcpy(&piloSpeedRefStr, &piloSpeedRefStrSave, sizeof(PILO_SPEED_REF_STR));
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT endTurn (PILO_TURN_STR *turnStr, int *bilan)
{
  memcpy(&piloSpeedRefStr, &piloSpeedRefStrSave, sizeof(PILO_SPEED_REF_STR));
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT endTraj (PILO_TRAJECTORY *smoothStr, int *bilan)
{
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT piloGotoConfigEnd(PILO_GOTO_CONFIG_STR *gotoStr, int *bilan)
{
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT endAbsSmooth (PILO_ABSOLUTE_BL *smoothGlobStr, int *bilan)
{
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT endStop (int *bilan)
{
  return(endExecTraj(bilan));
}
ACTIVITY_EVENT endTrajPoster (PILO_TRAJ_POSTER  *trajPoster, int *bilan)
{
  return(endExecTraj(bilan));
}


/*
 * Generation des point de consigne pour executer la trajectoire.
 *
 */
/*-------------- FONCTIONS AUXILIAIRES de genereTraj -----------------------*/


static ACTIVITY_EVENT genereTraj (int *bilan)
{
  static int numMvt;      /*XXXX*/
  int sens;
  static PILO_TYPE_MVT piloTypeMvt;
  double speed2;        /* Facteur de reduction de vitesse */
  double vReduce;       /* Vitesse reduite fonction du suivit du robot */
  double xref, yref, thetaref;

  /* Facteur de reduction de vitesse */
  piloSpeedRatio = (double)piloDataStrId->percentSpeed/100.;

  /* Initialisation */
  if (piloStateMvt == PILO_START_TRAJ) {
    numMvt = 0;
    piloStateMvt = PILO_SEG_START;
    piloTypeMvt = TYPE_MVT(0);
    piloDataStrId->refStr.numRef = 0;
    piloDataStrId->speedRef.numRef = 0;
    /*    piloDataStrId->refStr.nbPeriods = (int)
				      (piloCmdTaskPeriod/
				       LOCO_CMD_TASK_PERIOD + 0.1);*/
  }

  /* Test etats des serveurs */
  else {
    /* Le robot suit ? */
/*    vReduce = V_MAX(numMvt);
    waitForLoco(&vReduce);
    if (fabs(vReduce) > 0.0001 && fabs(V_MAX(numMvt)) > 0.0001) {
      speed2 = fabs(vReduce/V_MAX(numMvt));
      piloSpeedRatio = MIN(piloSpeedRatio2*piloSpeedRatio, piloSpeedRatio);
    }*/
  }

  /* Type de mouvement */
  switch(piloTypeMvt) {

    /* Execution des rotations */
  case TURN:
    piloStateMvt = piloTurnExec (numMvt, bilan);
    break;

    /* Execution des segments de droite */
  case SEGMENT:
    piloStateMvt = piloMoveExec (numMvt, bilan);
    break;

    /* Execution clothoide */
  case CLOTHO:
    piloStateMvt = piloClothExec (numMvt, bilan);
    break;

    /* Execution anti-clothoide */
  case ANTI_CLOTHO:
    piloStateMvt = piloAntiClothExec (numMvt, bilan);
    break;

    /* Trajectoire termine'e */
  case GOTO_END:
    /* Calculer la consigne pour le point en avant (ou en arriere) */
    /* Determination du sens du mouvement */
    sens = (piloDataStrId->refStr.backFlag == GENPOS_BACKWARD_MOTION) ? -1 : 1;
    piloDataStrId->refStr.x = X_START(numMvt)
      + sens * distPoint * cos (THETA_START(numMvt));
    piloDataStrId->refStr.y = Y_START(numMvt)
      + sens * distPoint * sin ( THETA_START(numMvt));
    piloDataStrId->refStr.theta = THETA_START(numMvt);
    piloDataStrId->refStr.v = 0.;
    piloDataStrId->refStr.w = 0.;

    piloDataStrId->speedRef.v = 0.;
    piloDataStrId->speedRef.w = 0.;

    piloRef.s = 0.;
    piloRef.ds = 0.;
    return(END);
  }


  /* Determination du sens du mouvement */
  sens = (piloRef.backFlag == GENPOS_BACKWARD_MOTION) ? -1 : 1;

  /* Calculer la consigne pour le point en avant (ou en arriere) */
  (piloDataStrId->refStr.numRef)++;
  (piloDataStrId->speedRef.numRef)++;

  piloDataStrId->refStr.x = piloRef.x + sens * distPoint * cos (piloRef.theta);
  piloDataStrId->refStr.y = piloRef.y + sens * distPoint * sin (piloRef.theta);
  piloDataStrId->refStr.theta = piloRef.theta;
  piloDataStrId->refStr.v = piloRef.v;
  piloDataStrId->refStr.w = piloRef.w;

  piloDataStrId->speedRef.v = PILO_SPEED_CORRECTION_V * piloRef.v;
  piloDataStrId->speedRef.w = PILO_SPEED_CORRECTION_W * piloRef.w;
  piloDataStrId->refStr.backFlag = piloRef.backFlag;

  GENOM_VERBOSE("GLOBAL x %5.2f y %5.2f th %5.2f v %5.2f w %5.2f",
	       piloDataStrId->refStr.x, piloDataStrId->refStr.y, piloDataStrId->refStr.theta,
	       piloDataStrId->refStr.v, piloDataStrId->refStr.w);

  /* convert to local */
  xref = (piloDataStrId->refStr.x - xtransf) * cos(thetatransf) +
	 (piloDataStrId->refStr.y - ytransf) * sin(thetatransf);
  yref = (piloDataStrId->refStr.y - ytransf) * cos(thetatransf) -
	 (piloDataStrId->refStr.x - xtransf) * sin(thetatransf);

  thetaref = piloDataStrId->refStr.theta - thetatransf;


  piloDataStrId->refStr.x = xref;
  piloDataStrId->refStr.y = yref;
  piloDataStrId->refStr.theta = thetaref;

  GENOM_VERBOSE("LOCAL x %5.2f y %5.2f th %5.2f v %5.2f w %5.2f\n",
	       piloDataStrId->refStr.x, piloDataStrId->refStr.y, piloDataStrId->refStr.theta,
	       piloDataStrId->refStr.v, piloDataStrId->refStr.w);

  /* Longueur parcourue pour le monitoring */
  piloRef.ds = fabs(piloRef.v) * piloCmdTaskPeriod;
  piloRef.s += piloRef.ds;

  /* Mouvement suivant */
  if (piloStateMvt == PILO_SEG_END) {
    numMvt++;
    piloTypeMvt = TYPE_MVT(numMvt);
    piloStateMvt = PILO_SEG_START;
  }
  if (piloStateMvt == ERROR) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }
  return(EXEC);
}

/*****************************************************************************
 *
 * endExecTraj - Termine l'execution d'une trajectoire
 *               (arret normal).
 *
 * Retourne: ETHER ou FAIL ou END
 */

static ACTIVITY_EVENT endExecTraj(int *bilan )
{

  /* Distances parcoures sur la traj (pour monitors) */
  piloRef.s = 0.;
  piloRef.ds = 0.;
  piloDataStrId->refStr.numRef = -1;
  piloDataStrId->speedRef.numRef = -1;

  piloDataStrId->refStr.v = 0.;
  piloDataStrId->refStr.w = 0.;
  piloDataStrId->speedRef.v = 0.;
  piloDataStrId->speedRef.w = 0.;

  return(ETHER);
}

/*****************************************************************************
 *
 * interExecTraj - Interrompt l'execution d'une trajectoire.
 *
 *
 * Retourne: ETHER ou FAIL
 */

static ACTIVITY_EVENT interExecTraj(int *bilan )
{
  /* Distances parcoures sur la traj (pour monitors) */
  piloRef.s = 0.;
  piloRef.ds = 0.;
  piloDataStrId->refStr.numRef = -1;
  piloDataStrId->speedRef.numRef = -1;
  piloDataStrId->refStr.v = 0.;
  piloDataStrId->refStr.w = 0.;
  piloDataStrId->speedRef.v = 0.;
  piloDataStrId->speedRef.w = 0.;

  return(ETHER);
}


static int piloTestGivenInitPos(double x0, double y0, double t0,
				double dt, double dl)
{
  GENPOS_CART_STATE pos;
  double deltaTh, deltaL;

  /* Current position */
  pos.xRob = x0;
  pos.yRob = y0;
  pos.theta = t0;
  if (getRobotPosition (&pos) == ERROR)
    return errnoGet();

  /* Deviation between theorical and current positions  */
  deltaL = sqrt(SQR(x0 - pos.xRob) + SQR(y0 - pos.yRob));
  deltaTh = angleLimit(t0 - pos.theta);

  /* Compatibility */
  /*  printf ("PILO: init\n"
	  "    (%6.3f %6.3f %6.3f) instead (%6.3f %6.3f %6.3f)\n"
	  "     DL %6.4f DTH %6.4f     v %6.4f w %6.4f\n",
	  pos.xRob, pos.yRob, pos.theta,
	  x0, y0, t0,
	  deltaL, deltaTh,
	  pos.v, pos.w);*/

  /* Test compatibility */
  if (fabs(deltaTh) > dt || deltaL > dl) {
    printf ("PILO: incompatible init. pos.\n");
    return S_pilo_INVALID_INITIAL_POSITION;
  }
  return OK;
}

/*****************************************************************************
 *
 * piloInitPos  -   Recuppere la config courante du robot et demarre le
 *                  tracking
 *
 * Retourne: OK ou bilan
 */

static int piloInitPos ()
{
  GENPOS_CART_STATE odo;

  /* Lecture de la position, vitesses, ... courants du robot */
  odo.xRob = X_START(0);
  odo.yRob = Y_START(0);
  odo.theta = THETA_START(0);
  if (getRobotPosition (&odo) == ERROR)
    return (errnoGet());
  X_START(0) = odo.xRob;
  Y_START(0) = odo.yRob;
  THETA_START(0) = odo.theta;
  V_START(0) = odo.v;
  W_START(0) = odo.w;

  return OK;
}

static int piloInitGivenPos (double x, double y, double th)
{
  X_START(0) = x;
  Y_START(0) = y;
  THETA_START(0) = th;

  V_START(0) = 0.;
  W_START(0) = 0.;

  return OK;
}


/*****************************************************************************
 *
 * piloInitMvt  -   Recuppere la config courante du robot et demarre le
 *                  tracking
 *
 * Retourne: OK ou bilan
 */

static int piloInitMvt (int nbMvts)
{

  /* Simulation de tracking avec depart en (0,0,0) */
/* odo.xRob = -distPoint;
   odo.yRob = 0;
   odo.theta = 0;
*/

  /* Initialiser le mouvement */
  piloStateMvt = PILO_START_TRAJ;

  /* Initialiser la consigne locale */
  piloRef.v = 0.;
  piloRef.w = 0.;
  piloRef.s = 0.;

  /* Enregistrer le nombre de mouvements */
  piloNbMvts = nbMvts;

  /* C'est OK */
  return (OK);
}


/******************************************************************************
 *
 *  piloTestAtFinalConfig
 *
 *  Description:
 *
 *  Retourne: TRUE FALSE
 */
static BOOL piloTestAtFinalConfig(int *bilan)
{
  GENPOS_CART_STATE odo;

  /* Lecture de la position courante du robot */
  odo.xRob = X_START(piloNbMvts-1);
  odo.yRob = Y_START(piloNbMvts-1);
  odo.theta = THETA_START(piloNbMvts-1);
  if (getRobotPosition (&odo) == ERROR) {
    *bilan = errnoGet();
    return FALSE;
  }

  /* Compare avec derniere consigne */
  if(TYPE_MVT(piloNbMvts-1) == GOTO_END) {
    if(fabs(odo.xRob - X_START(piloNbMvts-1)) < 0.05 &&
       fabs(odo.yRob - Y_START(piloNbMvts-1)) < 0.05 &&
       fabs(angleLimit(odo.theta - THETA_START(piloNbMvts-1)) < 0.05)) {
      printf ("PILO: at final config\n");
      *bilan = OK;
      return TRUE;
    }
  }
  else
    printf ("PILO: last motion is not GOTO_END\n");

  /* Bilan inchange */
  return FALSE;
}

/******************************************************************************
 *
 *  piloMoveExec  -  Calcul et execution des consignes en vu d'un mouvement de
 *                   type move.
 *
 *  Description:
 *  Cette routine calcul les consignes afin de parcourir une ligne droite de
 *  longueur d donnee a acceleration max tout en respectant les limites en
 *  vitesse et de facon a atteindre la vitesse finale en fin de consigne.
 *
 *  Retourne: ON en execution et END pour la fin d'exec
 */

static PILO_STATE_MVT piloMoveExec (int numMvt, int *bilan)

{
  static double distance=0.;            /* Distance parcourue */
  double epsDist;
  double distDecel;
  double vmax;
  double accel;
  double vFinale;
  double dDist;
  double accelRef;

  /* On initialise le mouvement */
  if (piloStateMvt == PILO_SEG_START) {

    if (ACCEL_LIN(numMvt) == 0.) {
      *bilan = S_pilo_ERROR_IN_DYNAMIC;
      return ERROR;
    }
    piloRef.w = 0.;
    distance = 0.;
  }

  /* Vitesse max reduite */
  vmax = piloSpeedRatio * V_MAX(numMvt);
  accelRef = MAX(fabs(ACCEL_LIN(numMvt)) * SQR(piloSpeedRatio),
		 ALIN_REF/PILO_REDUC_ACCEL_LIN);

  /* Vitesse finale reduite */
  vFinale = piloSpeedRatio * V_START(numMvt+1);

  /* Distance parcour pendant 1 periode */
  epsDist = MIN(fabs(piloRef.v) * piloCmdTaskPeriod,
		fabs(VAR_LIN(numMvt))/2.);
  epsDist = MAX(epsDist, EPSILON_DIST);

  /* Distance de deceleration */
  distDecel = fabs((SQR(piloRef.v) - SQR(vFinale))
	      /(2*accelRef))
		+ fabs(piloRef.v - vFinale) * piloCmdTaskPeriod/2.;

  /* On n'est pas encore en deceleration */
  if (fabs(VAR_LIN(numMvt)) - fabs(distance) > fabs(distDecel))  {

    /* Regulation autour de la vitesse max */
    accel = (vmax - piloRef.v)/piloCmdTaskPeriod;

    /* Limite l'acceleration */
    accel = SIGN(accel) * MIN(accelRef, fabs(accel));
    dDist = move (accel, &distance, numMvt);
  }

  /* Deceleration finale */
  else {
    /* Regulation vers la vitesse finale */
    accel = (vFinale - piloRef.v)/piloCmdTaskPeriod;

    /* Limite l'acceleration */
    accel = SIGN(accel) * MIN(accelRef, fabs(accel));
    dDist = move (accel, &distance, numMvt);

    /* On est arrive' */
    if (fabs(distance) >= fabs(VAR_LIN(numMvt)) - epsDist) {

      if (V_START(numMvt+1) == 0.) piloRef.v = 0.;
      piloRef.x = X_START(numMvt+1);
      piloRef.y = Y_START(numMvt+1);

      return(PILO_SEG_END);
    }
  }

  /* On ne bouge plus ! */
  if (dDist == 0. && accel == 0. && piloSpeedRatio != 0) {
    if (fabs(distance) >= fabs(VAR_LIN(numMvt)) - 0.1) {
      printf ("PILO move: motion without velocity !? Near goal (%f m): OK.\n",
	      VAR_LIN(numMvt)-distance);
      piloRef.x = X_START(numMvt+1);
      piloRef.y = Y_START(numMvt+1);
      /* Signaler la fin de l'execution */
      return(PILO_SEG_END);
    }
    else  {
      printf ("PILO move: motion without velocity !? Goal at %f m -> Bye.\n",
	      VAR_LIN(numMvt)-distance);
      *bilan = S_pilo_NUL_SPEED_MOTION;
      return ERROR;
    }
  }
  return(PILO_SEG_ON);
}



/******************************************************************************
*
*  move - Calcul des consignes pour ligne droite
*
*  Description: Calcul des consignes pour mouvements a acceleration angulaire
*               nulle par increment.

*
*  Retourne: la distance deja parcourue.
*/

static double move (double accel,            /* Acceleration lineaire signee */
		    double *distance,        /* Distance parcourue */
		    int numMvt)
{
  static double c, s;
  static double x0, y0;
  double dDist;

  if (piloStateMvt == PILO_SEG_START) {
    piloRef.w = 0.;
/*    piloRef.v = V_START(numMvt);*/
    piloRef.theta = THETA_START(numMvt);
    piloRef.backFlag = SIGN(VAR_LIN(numMvt)) < 0 ?
		       GENPOS_BACKWARD_MOTION : GENPOS_FORWARD_MOTION;
    c = cos(THETA_START(numMvt));
    s = sin(THETA_START(numMvt));
    x0 = X_START(numMvt);
    y0 = Y_START(numMvt);
  }

  dDist = piloRef.v *  piloCmdTaskPeriod;
  *distance += dDist + accel * SQR(piloCmdTaskPeriod)/2.;
  piloRef.v += accel * piloCmdTaskPeriod;

  piloRef.x = x0 + *distance * c;
  piloRef.y = y0 + *distance * s;

  return dDist;
}


/******************************************************************************
*
*  piloTurnExec  -  Calcul et execution des consignes en vu d'un mouvement de
*                   type turn.
*
*  Description:
*  Cette routine calcul les consignes afin de parcourir un arc de cercle de
*  rayon donne a acceleration max tout en respectant les limites en
*  vitesse et de facon a atteindre la vitesse finale en fin de consigne.
*
*  Retourne: ON en execution et END en fin d'execution
*/

static PILO_STATE_MVT piloTurnExec (int numMvt, int *bilan)

{
  double epsAng;
  double angDecel;
  static double parcour;
  double wmax;
  double accel;
  double wFinale;
  double dAng;
  double accelRef;

  /* Init. variation */
  if (piloStateMvt == PILO_SEG_START) {

    if (ACCEL_ANG(numMvt) == 0.
	|| (RO_START(numMvt) != 0. && ACCEL_LIN(numMvt) == 0.)) {
      printf ("PILO turn %d: aa %f r %f al %f \n",
	      numMvt, ACCEL_ANG(numMvt), RO_START(numMvt), ACCEL_LIN(numMvt));
      *bilan = S_pilo_ERROR_IN_DYNAMIC;
      return ERROR;
    }
    parcour = 0.;
  }

  /* Reduction de vitesse max */
  wmax = W_MAX(numMvt) * piloSpeedRatio;
  accelRef = MAX(fabs(ACCEL_ANG(numMvt)) * SQR(piloSpeedRatio),
		 AANG_REF/PILO_REDUC_ACCEL_ANG);

  /* Vitesse finale reduite */
  wFinale = piloSpeedRatio * W_START(numMvt+1);

  /* Angle de deceleration avec marge d'une periode */
  epsAng = MIN(fabs(piloRef.w) * piloCmdTaskPeriod,
	       fabs(VAR_ANG(numMvt))/2.);
  epsAng = MAX(epsAng, EPSILON_ANGLE);
  angDecel = fabs((SQR(piloRef.w) - SQR(wFinale)) /(2. * accelRef))
	     + fabs(piloRef.w - wFinale) * piloCmdTaskPeriod/2.
	     + 0.001;

  /* On n'est pas encore en deceleration */
  if (fabs(parcour) < fabs(VAR_ANG(numMvt)) - angDecel)  {

    /* Regulation autour de la vitesse max */
    accel = (wmax - piloRef.w)/piloCmdTaskPeriod;

    /* Limite l'acceleration angulaire */
    accel = SIGN(accel) * MIN(accelRef, fabs(accel));

    /* Limite l'acceleration lineaire */
/*    if (fabs (RO_START(numMvt)) > 0.0001)
      accel = SIGN(accel) * MIN(fabs(ACCEL_LIN(numMvt)/RO_START(numMvt)),
				 fabs(accel));*/

    dAng = turn (accel, RO_START(numMvt), &parcour, numMvt);
  }

  /* Deceleration finale */
  else {
    /* Regulation vers la vitesse finale */
    accel = (wFinale - piloRef.w)/piloCmdTaskPeriod;

    /* Limite l'acceleration angulaire */
    accel = SIGN(accel) * MIN(accelRef, fabs(accel));

    /* Limite l'acceleration lineaire */
/*    if (fabs (RO_START(numMvt)) > 0.0001)
      accel = SIGN(accel) * MIN(fabs(ACCEL_LIN(numMvt)/RO_START(numMvt)),
				 fabs(accel));*/


    dAng = turn (accel, RO_START(numMvt), &parcour, numMvt);

    /* On est arrive' */
    if (fabs(parcour) >= fabs(VAR_ANG(numMvt)) - epsAng) {
      if (W_START(numMvt+1) == 0.) {
	piloRef.v = 0.;
	piloRef.w = 0.;
      }
      piloRef.theta = THETA_START(numMvt+1);
      piloRef.x = X_START(numMvt+1);
      piloRef.y = Y_START(numMvt+1);

      /* Signaler la fin de l'execution */
      return(PILO_SEG_END);
    }
  }

  /* On ne bouge plus ! */
  if (dAng == 0. && accel == 0. && piloSpeedRatio != 0) {

    /* Ecart tolerable */
    if (fabs(parcour) >= fabs(VAR_ANG(numMvt)) - 0.08) {
      printf ("PILO turn: motion without velocity !? Near goal (%f rd): OK.\n",
	      VAR_ANG(numMvt)-parcour);
      piloRef.v = wFinale * RO_START(numMvt);
      piloRef.w = wFinale;
      piloRef.theta = THETA_START(numMvt+1);
      piloRef.x = X_START(numMvt+1);
      piloRef.y = Y_START(numMvt+1);

      /* Signaler la fin de l'execution */
      return(PILO_SEG_END);
    }
    /* Il faut trouver le bug */
    else {
      printf ("PILO turn: motion without velocity !? Goal at %f rd -> Bye.\n",
	      VAR_ANG(numMvt)-parcour);
      *bilan = S_pilo_NUL_SPEED_MOTION;
      return ERROR;
    }
  }

  return(PILO_SEG_ON);
}


/******************************************************************************
*
*  turn - Calcul des consignes pour arc de cercle
*
*  Description: Calcul des consignes pour mouvements circulaires par
*               increment.
*
*  Retourne: rien
*/

static double turn (double accelAng,         /* Acceleration lineaire signee */
		    double rayon,
		    double *angle,           /* Angle parcourue */
		    int numMvt)
{
  static double xRot, yRot;
  static double theta0;
  double c,s;
  double dAng;


  /* Initialisations */
  if (piloStateMvt == PILO_SEG_START) {
    /* Centre de rotation */
    theta0 = THETA_START(numMvt);
    c = cos(theta0);
    s = sin(theta0);
    xRot = X_START(numMvt) - rayon * s;
    yRot = Y_START(numMvt) + rayon * c;
/*    piloRef.w = W_START(numMvt);*/
    piloRef.backFlag = SIGN(VAR_ANG(numMvt) * rayon) < 0 ? GENPOS_BACKWARD_MOTION : GENPOS_FORWARD_MOTION;
    *angle=0.;
  }

  /* Angle atteint */
  dAng = piloRef.w * piloCmdTaskPeriod;
  *angle += dAng;

  /* Vitesses de consigne */
  piloRef.w += accelAng * piloCmdTaskPeriod;
  piloRef.v = piloRef.w * rayon;

  /* Coordonnees de consigne */
  piloRef.x = xRot + rayon * sin (*angle + theta0);
  piloRef.y = yRot - rayon * cos (*angle + theta0);

  /* Angle theorique atteint (pour asserv du point en avant) */
  piloRef.theta = theta0 + *angle;

  return dAng;
}


/*****************************************************************************
*
*  piloClothExec  -  Calcul et execution des virages
*
*  Description:
*
*  Retourne:
*/

static PILO_STATE_MVT piloClothExec (int numMvt, int *bilan)

{
  static double lin;         /* Distance parcourue */
  static double ang;
  static double distLin;       /* Variation angulaire totale */
  static double distAng;       /* Variation angulaire totale */
  double epsDist, epsAng;
  double vFinale;                /* Vitesse modifiee */

  /* On initialise le mouvement */
  if (piloStateMvt == PILO_SEG_START) {
    if (K_TRAJ(numMvt) == 0. || ACCEL_ANG(numMvt) == 0.) {
      *bilan = S_pilo_ERROR_IN_DYNAMIC;
      return ERROR;
    }

    lin = 0.;
    ang = 0.;
    distLin = fabs(VAR_LIN(numMvt));
    distAng = fabs(VAR_ANG(numMvt));
  }

  /* Vitesse finale/croisiere reduite */
  vFinale = piloSpeedRatio * V_START(numMvt+1);

  /* Distance parcouru pendant 1 periode */
  epsDist = MIN(fabs(piloRef.v) * piloCmdTaskPeriod,
		distLin/2.);
  epsDist = MAX(epsDist, EPSILON_DIST);

  /* Angle parcouru pendant 1 periode */
  epsAng = MIN(fabs(piloRef.w) * piloCmdTaskPeriod,
	       distAng/2.);
  epsAng = MAX(epsAng, EPSILON_ANGLE);

  /* Tant que la distance n'est pas atteinte */
  if ( fabs(lin) <= distLin - epsDist) {

    /* Calcul des consignes */
    clotho (vFinale, &lin, &ang, numMvt);
  }

  /* C'est fini */
  else {
    piloRef.x = X_START(numMvt+1);
    piloRef.y = Y_START(numMvt+1);
    piloRef.theta = THETA_START(numMvt+1);
    return(PILO_SEG_END);
  }

  return(PILO_SEG_ON);
}


/******************************************************************************
*
*  clotho - Calcul des consignes pour clotho
*
*  Description: Calcul des consignes pour mouvements de type clotho.
*  La geometrie de la clotho et les conditions initiales sont specifiees
*  dans la structure piloSmoothTrajId[numMvt].
*  La vitesse de parcour souhaitee est indiquee en parametre.
*  L'angle parcouru est incremente et retourne en parametre.
*
*  Retourne: ON, END
*/

static void clotho (double vitesse,   /* Vitesse lineaire souhaitee */
		    double *lin,      /* Distance lineaire parcourue */
		    double *ang,      /* Distance angulaire parcourue */
		    int numMvt)
{

  static double xCloth0, yCloth0;     /* Position init. dans le rep. clotho */
  static double x0, y0;               /* Position en debut de mouvement */
  static double kCloth;               /* Constante de la clotho
					 = v0 sqrt(PI/aa) = forw sqrt(pi/kc) */
  static double kCos0, kSin0;         /* Homothetie et rotation de la cloth */
  static double phi0;                 /* Angle initiale sur la clotho */
  static double leftFlag;             /* Flag rotation gauche/droite */
  static double signe;                /* Sens parcour de la clotho */

  static double courbure;             /* Courbure courante */
  static double distCloth;            /* Position curviligne sur la clotho */
  static double phi;                  /* Angle dans repere clotho */

  double accelLin;                    /* Acceleration lineaire signee */
  double theta0;                      /* Angle de rotation de la clotho */

  double xCloth, yCloth;              /* Coord dans le repere clotho */
  double dS, dPhi;                    /* Variation curviligne */

  double pCloth;                      /* Parametre courant de la clotho
					 pCloth = sqrt(2*phi/PI) , phi: angle
					 dans le repere clotho. */
  double dv, w;
  BOOL changeW;


  /* Initialisations */
  if (piloStateMvt == PILO_SEG_START) {

    /* Constante clotho = v0 sqrt(PI/aa) = forw sqrt(pi/kc) */
    kCloth = K_TRAJ(numMvt);

    /* Signe acceleration angulaire */
    leftFlag = SIGN(ACCEL_ANG(numMvt));

    /* Initialiser les consignes */
    piloRef.backFlag = SIGN(V_START(numMvt)) < 0 ?
		       GENPOS_BACKWARD_MOTION : GENPOS_FORWARD_MOTION;
    piloRef.theta = THETA_START(numMvt);

    /* Position initiale dans le repere clotho */
    courbure = C_START(numMvt);
    distCloth = leftFlag * courbure * SQR(kCloth) / PI;

    pCloth = leftFlag * kCloth * courbure / PI;
    signe = pCloth < 0. ? -1 : 1;
    xCloth0 = signe * integCosFresnel(fabs(pCloth));
    yCloth0 = signe * integSinFresnel(fabs(pCloth));
    phi0 = leftFlag * pCloth * pCloth * PI /2.;
    phi = phi0;

    /* Rotation de la clotho */
    theta0 = THETA_START(numMvt) - phi0;
    kCos0 = kCloth * cos(theta0);
    kSin0 = kCloth * sin(theta0);

    /* Position initiale */
    x0 = X_START(numMvt);
    y0 = Y_START(numMvt);
  }

  /* Regulation vers la vitesse souhaitee */
  dv = vitesse - piloRef.v;
  accelLin = fabs(dv)/piloCmdTaskPeriod;

  /* Limite l'acceleration angulaire */
  if (fabs(courbure) > 0.00001)
    accelLin = MIN(accelLin,
		    fabs((AMAX_ANG_PILO_MAJOR
			  - PI * SQR(piloRef.v/kCloth)) / courbure));

  /* Limite l'acceleration lineaire */
  accelLin = SIGN(dv) * MIN(accelLin, AMAX_LIN_PILO_MAJOR);

  /* Nouvelle vitesse lineaire */
  piloRef.v += accelLin * piloCmdTaskPeriod;
  dS = piloRef.v * piloCmdTaskPeriod - accelLin * SQR(piloCmdTaskPeriod)/2.;
  distCloth += dS;

  /* Courbure */
  courbure = leftFlag * PI / SQR(kCloth) * distCloth;

  /* Nouvelle vitesse angulaire */
  w = piloRef.v * courbure;

  /* Limitation de la vitesse angulaire */
  changeW = FALSE;
  if(fabs(w) > WMAX_PILO) {
    w = SIGN(w)*WMAX_PILO;
    changeW = TRUE;
  }
  if(fabs(piloRef.w-w)>AMAX_ANG_PILO_MAJOR*piloCmdTaskPeriod) {
    w = piloRef.w + SIGN(w-piloRef.w)*AMAX_ANG_PILO_MAJOR*piloCmdTaskPeriod;
    changeW = TRUE;
  }
  if(changeW) {
    dPhi = w * piloCmdTaskPeriod;
    phi += dPhi;
    piloRef.theta += dPhi;
    pCloth = signe * sqrt(2.*fabs(phi)/PI);
    distCloth = pCloth * kCloth;
    courbure = leftFlag * PI / SQR(kCloth) * distCloth;
    piloRef.v = w/courbure;
    dS = piloRef.v * piloCmdTaskPeriod;
  }

  else {
    /* Parametre courant de la clotho */
    pCloth = distCloth/kCloth;

    /* Angle sur la clotho */
    phi = leftFlag * PI * SQR(pCloth) / 2.;
    piloRef.theta = THETA_START(numMvt) + phi - phi0;
    dPhi = w * piloCmdTaskPeriod;
  }

  piloRef.w = w;

  /* Distance parcourue */
  *lin += dS;
  *ang += dPhi;

  /* Variations normalise'es en x et y dans rep. clotho */
  xCloth = signe * integCosFresnel(fabs(pCloth)) - xCloth0;
  yCloth = leftFlag * (signe * integSinFresnel(fabs(pCloth)) - yCloth0);

  /* Coordonnees */
  piloRef.x =  x0 + kCos0 * xCloth - kSin0 * yCloth;
  piloRef.y =  y0 + kSin0 * xCloth + kCos0 * yCloth;
}



/*****************************************************************************
*
*  piloAntiClothExec  -  Calcul et execution des virages
*
*  Description:
*
*
*  Retourne:  rien
*/

static PILO_STATE_MVT piloAntiClothExec (int numMvt, int *bilan)

{
  static double parcour;         /* Angle parcouru */
  static double alpha;           /* Variation angulaire totale virage */
  static double epsAng;          /* Discretisation angulaire */
  double vFinale, wFinale;       /* Vitesse finale souhaitee */

  /* On initialise le mouvement */
  if (piloStateMvt == PILO_SEG_START) {
    if (K_TRAJ(numMvt) == 0. || ACCEL_LIN(numMvt) == 0.) {
      *bilan = S_pilo_ERROR_IN_DYNAMIC;
      return ERROR;
    }
    alpha = fabs(VAR_ANG(numMvt));
    parcour = 0.;
  }

  /* Vitesses finales reduites */
  vFinale = piloSpeedRatio * V_START(numMvt+1);
  if (fabs(RO_START(numMvt+1)) > 0.00001)
    wFinale = vFinale/RO_START(numMvt+1);
  else
    wFinale = 0.;

  /* Angle parcouru pendant 1 periode */
  epsAng = MIN(fabs(piloRef.w) * piloCmdTaskPeriod,
	       fabs(VAR_ANG(numMvt)/2.));
  epsAng = MAX(epsAng, EPSILON_ANGLE);

  /* Tant que l'angle n'est pas atteint */
  if (fabs(parcour) < alpha - epsAng) {

    /* Calcul des consignes */
    antiClotho(wFinale,  &parcour, numMvt);
  }

  /* C'est fini */
  else {
    piloRef.x = X_START(numMvt+1);
    piloRef.y = Y_START(numMvt+1);
    piloRef.theta = THETA_START(numMvt+1);
    return(PILO_SEG_END);
  }
  return(PILO_SEG_ON);
}

/*****************************************************************************
*
*  antiClotho - Calcul des consignes pour antiClotho
*
*  Description: Calcul des consignes pour mouvements de type anti clotho.
*
*
*  Retourne: ON, END
*/

static void antiClotho (double vitesse,        /* Vitesse souhaitee */
			double *angle,         /* Angle parcourue */
			int numMvt)
{
  static double x0, y0;            /* Position initiale */
  static double cosTh0, sinTh0;    /* Position initiale */
  double cosTh, sinTh;
  static double dAngle;
  static double phi;               /* Angle dans le repere anti-clotho */
  static double kAnti;             /* Constante anti-cloth */
  static double ro;                /* Rayon de gyration */
  static double ro0;               /* Rayon de gyration initial */
  double accelAng;
  double dw;

  /* Initialisations */
  if (piloStateMvt == PILO_SEG_START) {
    kAnti = K_TRAJ(numMvt);

    /* Configuration initiale */
    x0 = X_START(numMvt);
    y0 = Y_START(numMvt);
    ro0 = RO_START(numMvt);
    ro = ro0;
    phi = ro/kAnti;
    piloRef.theta = THETA_START(numMvt);
    sinTh0 = sin(THETA_START(numMvt));
    cosTh0 = cos(THETA_START(numMvt));
  }

  /* Regulation vers la vitesse souhaitee */
  dw = vitesse - piloRef.w;
  accelAng = fabs(dw)/piloCmdTaskPeriod;

  /* Limite l'acceleration angulaire */
  accelAng = MIN(accelAng, AMAX_ANG_PILO_MAJOR);

  /* Limite l'acceleration lineaire */
  if (fabs(ro) > 0.0001)
    accelAng = MIN(accelAng,
		    fabs(AMAX_LIN_PILO_MAJOR - kAnti * SQR(piloRef.w)/ro));
  accelAng *= SIGN(dw);

  /* Nouvelle vitesse angulaire */
  piloRef.w += accelAng * piloCmdTaskPeriod;
  dAngle = piloRef.w * piloCmdTaskPeriod + accelAng * SQR(piloCmdTaskPeriod)/2.;
  *angle += dAngle;
  phi += dAngle;

  /* Vitesse lineaire */
  ro = kAnti * phi;
  piloRef.v = ro * piloRef.w;

  /* Angle robot */
  piloRef.theta += dAngle;
  cosTh = cos(piloRef.theta);
  sinTh = sin(piloRef.theta);

  /* Coordonnees */
  piloRef.x = x0 + kAnti * (cosTh - cosTh0)
    + (ro * sinTh - ro0 * sinTh0);
  piloRef.y = y0 + kAnti * (sinTh - sinTh0)
    - (ro * cosTh - ro0 * cosTh0);
  piloRef.backFlag = SIGN(piloRef.v + 0.000001) < 0 ?
		     GENPOS_BACKWARD_MOTION : GENPOS_FORWARD_MOTION;

}


/******************************************************************************
 *
 *  piloMvtsFreeAndAlloc  -  Allocation de structures elementaires de mouvement
 *
 *
 *  Description:
 *
 *  Retourne:  rien
 */

static int piloMvtsFreeAndAlloc (int nbMvts)

{

  if (piloSmoothTrajId != NULL)
    free ((void*)piloSmoothTrajId);

  if (nbMvts==0)
    return ERROR;

  if ((piloSmoothTrajId = (SMOOTH_TRAJ *) malloc (nbMvts * sizeof(SMOOTH_TRAJ)))
      == NULL) {
    printf ("pilo: PB MALLOC\n");
    return (errnoGet());
  }
  memset((void *)piloSmoothTrajId, 0, nbMvts * sizeof(SMOOTH_TRAJ));

  return(OK);
}

/*****************************************************************************
 *
 *  Fonctions d'execution.
 *
 *  Parametres: pointeurs sur : Input, Output, SDI, bilan
 *
 *  Retourne: END, FAIL ou EXEC
 */


/*
 *  Reduit la vitesse si le robot ne suit pas
 *
 */

static int waitForLoco (double *vRef)
{
  GENPOS_CART_STATE odo;
  double vRp;              /* Vitesse robot projetee sur la traj */
  double cosDTheta;           /* Ecart angulaire robot-consigne */
  double vOffset=0.03;     /* Offset pour continuer a avancer */
  double dist;
  double distWait=0.1;     /* Distance avant regulation en vitesse */
/*  double distReduce = 1.; */ /* Distance pour regulation en distance */
  double dLong;            /* Ecart transversal */
  GENPOS_CART_REF * ref = &(piloDataStrId->refStr);


  /* Lecture de la position, vitesses, ... courants du robot */
  if (getRobotPosition (&odo) == ERROR)
    return (errnoGet());

  /* Ecart angle de consigne et angle vrai */
  cosDTheta = cos(ref->theta - odo.theta);

  /* Ecart position-consigne */
  dist = sqrt(SQR(ref->x - odo.xRef) + SQR(ref->y - odo.yRef));

  /* Ecart longitudinal */
  dLong = (ref->x - odo.xRef) * cos(ref->theta)
    + (ref->y - odo.yRef) * sin(ref->theta);

  /* Vitesse de consigne non negligeable,
     ET ecart position consigne-vrai non negligeable */
  if (fabs(ref->v) > vOffset && dist > distWait) {

    /* La consigne et le deplacement ne sont pas de meme sens */
    if (odo.v * ref->v < 0 || cosDTheta < 0.)
      vRp = 0.;

    /* Projection de  la vitesse sur la trajectoire */
    else
      vRp = odo.v * cosDTheta;

    /* Regulation en vitesse  */
/*    *vRef = (fabs(vRp) + vOffset) * SIGN(ref->v);*/

    /* Regulation en distance */
    *vRef =  SIGN(ref->v) *
      MAX( fabs(ref->v * exp(-2.*fabs(dLong))/0.13533528),
	  vOffset);
    return(OK);
  }

  return(OK);
}

void piloPrintPiloTraj(FILE *pfile)
{
  int i;
  char tabType[] = "mtCAE";
  i = 0;

  fprintf(pfile, "num\tseg\ttype\tdist\tang\t k\t v\t w\t al\t aa\n");
  while (TYPE_MVT(i) != GOTO_END) {
    fprintf (pfile, "%d: pos %5.2f %5.2f %5.2f\n"
	     "\t%d\t%c\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\n",
	     i, X_START(i), Y_START(i), THETA_START(i),
	     SEG_NUM(i), tabType[TYPE_MVT(i)], VAR_LIN(i), VAR_ANG(i),
	    K_TRAJ(i),
	    V_START(i), W_START(i), ACCEL_LIN(i), ACCEL_ANG(i));
    i++;
  }
  fprintf (pfile, "%d: pos %5.2f %5.2f %5.2f\n"
	   "\t%d\t%c\n",
	  i, X_START(i), Y_START(i), THETA_START(i),
	   SEG_NUM(i), tabType[TYPE_MVT(i)]);
}

void piloPrintInputTraj(FILE *pfile)
{
  static int piloNumTraj = 0;
  int i;

  fprintf (pfile, "num %d  init %6.2f %6.2f %6.2f   %d  ",
	   piloNumTraj, X_START(0), Y_START(0), THETA_START(0),
	   DATA_NB);

  /* Move, turn ou stop */
  if (TYPE_MVT(1) == GOTO_END) {
    if (TYPE_MVT(0) == SEGMENT)
      fprintf (pfile, "SEGMENT\n");
    else if (TYPE_MVT(0) == TURN)
      fprintf (pfile, "ARC\n");

    fprintf (pfile, " dth %6.2f  dl %6.2f\n",
	     VAR_ANG(0), VAR_LIN(0));
    return;
  }

  /* Ligne brisee */
  else if (DATA_TYPE == PILO_LINE_SEG)
    fprintf (pfile, "BROKEN LINE\n");

  /* R et S */
  else
    fprintf (pfile, "ARCS LINE\n");

  /* Trajectoire */
  for (i=0; i<DATA_NB; i++) {
    fprintf (pfile, " dth %6.2f  dl %6.2f\n",
	     DATA_DTH(i), DATA_DL(i));
  }
  return;
}

#ifdef SAVE_TRAJ
void piloSaveIO()
{
  static int fileNum=0;
  char yesChar;
  int notOk=TRUE;
  char baseFileName[128];
  char fileName[128];
  FILE *pfile;

 /* Mettre la bufferisation a zero char */
  setbuf (stdin, NULL);
  setbuf (stdout, NULL);

  /* Record in file */
  while (notOk) {
    printf ("FileName (or n) : ");
    h2scanf("%s", baseFileName);
    if (baseFileName[0] == 'n' && baseFileName[0] == '\n')
      notOk=FALSE;
    else {
      sprintf(fileName, "%s.%03d", baseFileName, fileNum);
      if ((pfile = fopen(fileName, "w")) == NULL) {
	printf ("Cannot open %s\n", fileName);
	printf ("Bye\n");
	return;
      }
      else {
	if (fprintf(pfile, "test %d\n", fileNum) <= 0) {
	  printf ("Cannot write %s\n", fileName);
	  printf ("Bye\n");
	  return;
	}
	printf ("Open %s ok\n", fileName);
	notOk=FALSE;
	fileNum++;
      }
    }
  }

  /* loop for several acquisitions */
  do {

    piloPrintInputTraj(pfile);
    piloPrintPiloTraj(pfile);
    fclose(pfile);

    /* Go on ? */
    printf ("\n\nAgain (new file) (y/n) ? [y] ");
    yesChar = 'y';
    h2scanf("%c", &yesChar);
    if (yesChar == 'n') {
      printf("bye\n");
      return;
    }
    sprintf(fileName, "%s.%03d", baseFileName, fileNum);
    if ((pfile = fopen(fileName, "w")) == NULL)
      printf ("Cannot open %s\n", fileName);
    else {
      fileNum++;
    }
  } while(1);
}
#endif /* SAVE_TRAJ */

static int piloWritePosterTraj(BOOL trajFlag)
{
  static PILO_TRAJECTORY traj;
  static int piloNumTraj = 0;
  double size;
  int i;

  if (GENOM_VERBOSE_LEVEL>=1) {
    piloPrintInputTraj(stdout);
  }

  traj.header.x0     = X_START(0);
  traj.header.y0     = Y_START(0);
  traj.header.theta0 = THETA_START(0);
  traj.header.numTraj = piloNumTraj++;

  /* Trajectoire */
  if(trajFlag) {
    traj.header.segType = DATA_TYPE;
    traj.header.nbSegs = DATA_NB;
    for (i=0; i<DATA_NB; i++) {
      traj.seg[i].dTheta = DATA_DTH(i);
      traj.seg[i].dL = DATA_DL(i);
    }
  }

  /* Move, Turn, Stop */
  else {
    traj.header.nbSegs = 1;
    traj.seg[0].dTheta = VAR_ANG(0);
    traj.seg[0].dL = VAR_LIN(0);
  }

  /* Sauvegarde dans le poster */
  size = sizeof(PILO_TRAJ_HEADER) + traj.header.nbSegs * sizeof(PILO_SEGMENT);
  if (posterWrite(PILO_BASICTRAJ_POSTER_ID, 0, (char *) &traj, size) != size)
    return errnoGet();

  return OK;
}





/* ============================== LOCAL FUNC ============================== */

/*------------------------------------------------------------------------
 * getRobotPosition
 *
 * Description:
 *
 * Bilans:      OK
 */


static BOOL getRobotPosition (GENPOS_CART_STATE *pos)
{
  GENPOS_CART_STATE posLocal;
  int sens;
  POM_POS pomPos ;
  POM_ME_POS pomMEPos;

  if (!posPosterId) {
    thetatransf = 0;
    xtransf = 0;
    ytransf = 0;
    return FALSE;
  }

  /* lectude de la position */
  if (pomMEPosPosterType) {
    if (piloPOM_ME_POSPosterRead(posPosterId, &pomMEPos) != OK) {
      printf("pilo: getRobotPosition: can't read posPoster\n");
      posPosterId = NULL;
      return FALSE;
    }
    pos->xRob = pomMEPos.main.euler.x;
    pos->yRob = pomMEPos.main.euler.y;
    pos->theta =  pomMEPos.main.euler.yaw;
    thetatransf = 0.0;
    xtransf = 0.0;
    ytransf = 0.0;
  } else {
    if (piloPOM_POSPosterRead(posPosterId, &pomPos) != OK) {
      printf("piloExecInit: can't read posPoster\n");
      posPosterId = NULL;
      return FALSE;
    }
    /* Position correction */
    pos->xRob = pomPos.mainToOrigin.euler.x;
    pos->yRob = pomPos.mainToOrigin.euler.y;
    pos->theta = pomPos.mainToOrigin.euler.yaw;

    posLocal.xRob = pomPos.mainToBase.euler.x;
    posLocal.yRob = pomPos.mainToBase.euler.y;
    posLocal.theta = pomPos.mainToBase.euler.yaw;
    
    thetatransf = pos->theta - posLocal.theta;
    xtransf = pos->xRob - (posLocal.xRob * cos(thetatransf) - posLocal.yRob * sin(thetatransf));
    ytransf = pos->yRob - (posLocal.xRob * sin(thetatransf) + posLocal.yRob * cos(thetatransf));
  }


  sens = (piloRef.backFlag == GENPOS_BACKWARD_MOTION) ? -1 : 1;

  pos->xRef =  pos->xRob  + sens * distPoint * cos (pos->theta);
  pos->yRef =  pos->yRob  + sens * distPoint * sin (pos->theta);

  GENOM_VERBOSE("getRobotPosition %.2lf %.2lf %.2lf",
		 pos->xRob, pos->yRob, pos->theta);


  return TRUE;
}


/*------------------------------------------------------------------------
 * GotoConfig
 *
 * Description:
 *
 * Bilans:      OK
 */

/* piloGotoConfigStart  -  codel START of GotoConfig
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
piloGotoConfigStart(PILO_GOTO_CONFIG_STR *gotoStr, int *bilan)
{
  int nbMvts, numMvt;
  double dx, dy, dtheta;

  /* Check arguments validity */
  if (gotoStr->speedPercent < 1 || gotoStr->speedPercent > 100) {
    *bilan = S_pilo_bad_parameters;
    return ETHER;
  }

  /* Alloc array of elementary motions */
  switch (gotoStr->gotoConfigMode) {
  case PILO_SPIN_TOWARD_POSITION:
    nbMvts = 2;
    break;
  case PILO_GOTO_POSITION:
    nbMvts = 3;
    break;
  case PILO_GOTO_CONFIG:
    nbMvts = 4;
    break;
  default:
    nbMvts = 0;
  }
  if ((*bilan = piloMvtsFreeAndAlloc (nbMvts)) != OK)
    return ETHER;

  /* Record initial position */
  if ((*bilan = piloInitPos ()) != OK)
    return ETHER;

  if (gotoStr->absOrRelConfig == PILO_ABSOLUTE_TRAJ) {
    dx = gotoStr->x - X_START(0);
    dy = gotoStr->y - Y_START(0);
    dtheta = atan2 (dy, dx);
    dtheta = angleLimit(dtheta - THETA_START(0));
  }
  else {
    dx = gotoStr->x;
    dy = gotoStr->y;
    dtheta = atan2 (dy, dx);
    dtheta = angleLimit(dtheta);
  }

  /* Record initial spin */
  numMvt = 0;
  TYPE_MVT(numMvt) = TURN;
  VAR_ANG(numMvt) = dtheta;
  VAR_LIN(numMvt) = 0.;
  RO_START(numMvt) = 0.;
  C_START (numMvt) = 1.E10;
  W_START(numMvt) = 0.;
  V_START(numMvt) = 0.;

  /* end of spin */
  X_START(numMvt+1) = X_START(numMvt);
  Y_START(numMvt+1) = Y_START(numMvt);
  THETA_START(numMvt+1) = angleLimit(THETA_START(numMvt) + dtheta);
  RO_START(numMvt+1) = 0.;
  C_START(numMvt+1)  = 1.E10;
  V_START(numMvt+1)  = 0.;
  W_START(numMvt+1)  = 0.;

  /* following move */
  if (gotoStr->gotoConfigMode != PILO_SPIN_TOWARD_POSITION) {
    numMvt++;
    TYPE_MVT(numMvt)  = SEGMENT;
    VAR_LIN(numMvt) = sqrt(dx*dx+ dy*dy);
    RO_START(numMvt) = 1.E10;
    C_START(numMvt) = 0.;
    V_START(numMvt) = 0.;
    W_START(numMvt) = 0.;

    /* end of move */
    X_START(numMvt+1) = X_START(numMvt) + VAR_LIN(numMvt) * cos(THETA_START(numMvt));
    Y_START(numMvt+1) = Y_START(numMvt) + VAR_LIN(numMvt) * sin(THETA_START(numMvt));
    THETA_START(numMvt+1) = THETA_START(numMvt);
    RO_START(numMvt+1) = 1.E10;
    C_START(numMvt+1)  = 0.;
    V_START(numMvt+1)  = 0.;
    W_START(numMvt+1)  = 0.;

    /* last spin */
    if (gotoStr->gotoConfigMode == PILO_GOTO_CONFIG) {
	numMvt++;
	if (gotoStr->absOrRelConfig == PILO_ABSOLUTE_TRAJ)
	  dtheta = angleLimit(gotoStr->theta - THETA_START(numMvt));
	else
	  dtheta = angleLimit(gotoStr->theta - THETA_START(0) - THETA_START(numMvt));
	TYPE_MVT(numMvt) = TURN;
	VAR_ANG(numMvt) = dtheta;
	VAR_LIN(numMvt) = 0.;
	RO_START(numMvt) = 0.;
	C_START (numMvt) = 1.E10;
	W_START(numMvt) = 0.;
	V_START(numMvt) = 0.;

	/* end of spin */
	X_START(numMvt+1) = X_START(numMvt);
	Y_START(numMvt+1) = Y_START(numMvt);
	THETA_START(numMvt+1) = angleLimit(THETA_START(numMvt) + dtheta);
	RO_START(numMvt+1) = 1.E10;
	C_START(numMvt+1)  = 0.;
	V_START(numMvt+1)  = 0.;
	W_START(numMvt+1)  = 0.;
    }
  }

  /* Point d'arrivee */
  numMvt++;
  TYPE_MVT(numMvt) = GOTO_END;

  /* Calcul des vitesses */
  memcpy(&piloSpeedRefStrSave, &piloSpeedRefStr, sizeof(PILO_SPEED_REF_STR));
  VMAX_REF = VMAX_PILO * (double) gotoStr->speedPercent/100.;
  WMAX_REF = WMAX_PILO * (double) gotoStr->speedPercent/100.;
  ALIN_REF = AMAX_LIN_PILO * (double) gotoStr->accelPercent/100.;
  AANG_REF = AMAX_ANG_PILO * (double) gotoStr->accelPercent/100.;
  GENOM_VERBOSE("Valeur de Vmax et Alin : %f %f\n", VMAX_REF, ALIN_REF);

  /* Determination of speed profils along the trajectory */
  if ((*bilan = piloSpeedProfiles (nbMvts)) != OK) return ETHER;

  /* record traj in poster */
  if ((*bilan = piloWritePosterTraj(FALSE)) != OK)
    return ETHER;

  /* Init traj */
  if ((*bilan = piloInitMvt(nbMvts)) != OK)
    return ETHER;

  return EXEC;
}
